# pests.r
#
# A PESTS source for R  (and maybe S-plus)
#
# PESTS = Poisson Estimators for State-space Time Series
# VERSION 1.0
#
# Patrick Brandt
# brandt@unt.edu
# Department of Political Science
# University of North Texas
#
# This file contains no warranty or guarantee of performance.
# Use at your own risk for non-commercial use
#
# Contents based on
#
# Brandt et al. 2000. "Dynamic Modelling for Persistent Event
# Count Time Series." American Journal of Political Science. 44(4):
# 823-843.
#
# Brandt and Williams. 2001. "A Linear Poisson Autoregressive
# Model: The PAR(p)"  Political Analysis 9(2).
#
# IF YOU USE THESE FUNCTIONS, CITE THE ABOVE REFERENCES.
#
# These functions are translated with modifications from the original
# GAUSS code.
# Copyright 1998-2000, Patrick T. Brandt and John T. Williams.
#
# R version of the code copyright 2001, Patrick T. Brandt.
#
#--------------------------------------------------------------#

# Get the time series library and some utility functions

if (exists("is.R") && is.function(is.R) && is.R()) { 
library(ts)
}

# EWMA function for use in the PEWMA model
# Computes a simple EWMA
ewma<-function(x,omega)
  { ind<-omega^(rev(seq(1:length(x)))) 
    wm<-weighted.mean(x,ind)
    wm
  }

# PEWMA code
# There are 4 functions:
# 1) DGP function
# 2) PEWMA Kalman filter
# 3) PEWMA model log-likelihood function
# 4) PEWMA wrapper that estimates a PEWMA regression model using optim()

# pewmadgp -- PEWMA data generation process
# n = number of observations in generated series
# a0 = initial value of a0, shape parameter for gamma prior
# b0 = initial value of b0, scale parameter for gamma prior
# w = omega, EWMA weight parameter
# x = matrix of regressors, NOT including an intercept, unless you
#     want a model with a growth term.
# d = delta, vector of regression coefficients for x.
#

pewmadgp<-function(n,a0,b0,w,x,d=c(0))
  { y<-matrix(0,nrow=n); mu<-matrix(0,nrow=n);
    d<-matrix(d,ncol=1)
    r<-matrix(0, nrow=n, ncol=1);
    for(t in 1:n)
    {   if(t==1)
          { r[t]<-digamma(a0) - digamma(w*a0)
            m<-(a0*w*exp(r[t]))/(b0*w*exp(-x[t,]%*%d))
            y[t]<-rpois(1,m)
            # update to compute posterior
            a<-a0*w + y[t]
            b<-b0*w + exp(x[t,]%*%d + r[t]);
            mu[t]<-y[t]
          }
        else
          { r[t]<-digamma(a) - digamma(w*a)
            aa<-a*w
            bb <- b*w*exp(-x[t,]%*%d - r[t])
            m<-(exp(x[t,]%*%d + r[t])*(ewma(y[1:t-1],w))/(ewma((exp(x[1:t-1,]%*%d + r[1:t-1])),w)))
            y[t]<-rpois(1,m)
            rm(m)
            a<-aa + y[t]
            b<-w*b + exp(x[t,]%*%d + r[t]);
            mu[t]<-exp(x[t,]%*%d + r[t])*a/b;
          }
      }
    y<-ts(y,start=1)
    y
  }

# pewmafilter -- Extended Kalman filter for PEWMA model.  This is used
# to compute the filter and the parameters for the log-likelihood
# function, pewmallf().

pewmafilter<-function(y,x,w,d)
{  # Initialize some matrices we need to store the filter
   a<-matrix(0,nrow=length(y)); aa<-a; aaa<-a;
   b<-a; bb<-a; bbb<-a;
   mu<-a; r<-a;
   # Loop over the filter calculations
  for(t in 1:length(y))
    { if(t==1)
        { aaa[t]<-y[t]; bbb[t]<-exp(x[t,]%*%d + r[t,1]);
          mu[t]<-mean(y);
        }
      else
        { a[t]<-aaa[t-1];  b[t,1]<-bbb[t-1];
          r[t]<-digamma(a[t]) - digamma(w*a[t]);
          aa[t]<-w*a[t,1];
          bb[t]<-w*b[t,1]*exp(-x[t,]%*%d)*exp(-r[t]);
          aaa[t]<-w*a[t,1] + y[t];
          bbb[t]<-w*b[t,1] + exp(x[t,]%*%d + r[t]);
          mu[t]<-exp(x[t,]%*%d + r[t])*aaa[t]/bbb[t];
        }
    }
   # Clean up and store the filter components in a ts object
   dta<-matrix(cbind(a,aa,aaa,b,bb,bbb,mu,r),ncol=8)
   dta<-ts(dta, start=1, names=c("a","aa","aaa","b","bb","bbb","mu","r"))
   dta
 }

# pewmallf -- PEWMA log-likelihood function.
# Returns the value of the LLF for the given data and parameters.

pewmallf<-function(p,y=y,x=matrix(0,nrow=length(y)))
  { # initialize the parameters
    if(length(p)==1)
      { w<-p[1];  d<-as.matrix(0) }
    else
      { w<-p[1]; d<-as.matrix(p[2:length(p)]) }
    y<-as.ts(y); x<-as.ts(x);
    # Call the filter
    pf<-pewmafilter(y,x,w,d);
    # Extract vectors of parameters from filter output
    pf<-pf[2:nrow(pf),];
    a<-pf[,"a"]; b<-pf[,"b"]; r<-pf[,"r"];
    x<-x[2:nrow(x),];    y<-y[2:nrow(y)];
    # Compute log-likelihood function
    lp<-(lgamma((w*a) + y) - lgamma(w*a)
         - lgamma(y+1) + ((w*a) * log(w*b*exp(-x%*%d - r)))
         - ((w*a) + y)*(log(1 + (w*b*exp(-x%*%d - r)))));
    lp<-sum(lp)
    lp
  }

# Pewma -- this is the main function for calling the PEWMA model
# estimation.
# y = Event count time series you wish to fit.  (can be a matrix or a
# ts object)
# x = matrix of regressors, not including an intercept, unless you
# want a growth term.
# init.param = starting values for the optimizer.  Use a glm() call to
# compute these.
#
# Note that the R version uses the constrained optimizer for the
# parameter w.  This is because 0 < w <= 1.  The parameter can go
# slightly over the bound of 1 because this is where the Pewma reduces
# to the Poisson model.  If the bound bites for d, just alter the
# value of 100000 to something else.

Pewma<-function(y,x=matrix(0,nrow=length(y)),init.param=c(1),...)
  { y<-matrix(y,ncol=1)
    # Set up some bounds for the "omega" parameter of the PEWMA model
    k <- length(init.param)
    lower.bound <- c(0,rep(-100000,k-1))
    upper.bound <- c(1.1,rep(100000,k-1))
    if (exists("is.R") && is.function(is.R) && is.R())
      { fitted.param<-optim(init.param, pewmallf, method = "L-BFGS-B",
                            lower=lower.bound, upper=upper.bound,
                            hessian=TRUE, 
                            control=list(trace=2, REPORT=5, fnscale=-(length(y))),
                            y=y,x=x,...)
        param<-fitted.param$par
        llf <- fitted.param$value

      }
    else
      {
      fitted.param<-nlminb(start = init.param, obj = pewmallf,
                           hessian=F, scale = -(length(y)), y=y, x=x, ...)
      param<-fitted.param$parameters
      } 
    se<-sqrt(diag(-solve(fitted.param$hessian)))
    z<-param/se
    coefs <- t(matrix(rbind(param,se,z),nrow=3));
    colnames(coefs) <- c("Parameters","Std. Errors","Z-score")
    rownames(coefs) <- colnames(x)
     cat(" ","\n")
     cat("PEWMA regression output","\n")
     cat("--------------------------------------","\n")
     print.coefmat(coefs)
     cat("--------------------------------------","\n")
    fit <- list(coefs=coefs,
                   optim.param=fitted.param,
                   param=param,
                   std.err=se,
                   z=z,
                   llf=llf)
    class(fit) <- c("pests")
    invisible(fit)
  }


# PAR(p) code
# There are 4 functions:
# 1) DGP function
# 2) PAR(p) Kalman filter
# 3) PAR(p) model log-likelihood function
# 4) PAR(p) wrapper that estimates a PAR(p) regression model using optim()

# parpdgp -- PAR(p) data generation process
# n = number of observations in generated series
# r = AR(p) coefficients
# x = matrix of regressors, including an intercept.
# d = delta, vector of regression coefficients for x.
#

parpdgp<-function(n,r,x=matrix(0,nrow=n),d=as.matrix(0))
  { # Initialize the parameters
    y<-matrix(0,nrow=n)
    r<-as.matrix(r)
    p<-nrow(r)
    xd<-exp(as.matrix(as.matrix(x)%*%as.vector(d)))
    # Loop over DGP
    for(t in 1:n)
      { if(t<(p+1))
          { y[t]<-rpois(1,xd[t]) }
        else
          { ylag<-t(y[(t-p):(t-1)])%*%rev(r)
            y[t]<-rpois(1,(ylag + (1-sum(r))*xd[t]))
          }
      }
    return(ts(y,start=1))
  }


# parpfilter -- Extended Kalman filter for PAR(p) model.  This is used
# to compute the filter and the parameters for the log-likelihood
# function, parpllf().

parpfilter<-function(y,x,r,d)
  { # Initialize some parameters
    y<-as.matrix(y)
    r<-as.matrix(r)
    p<-nrow(r)
    n <- length(y)
    xd<-exp(as.matrix(as.matrix(x)%*%as.vector(d)))
    m<-matrix(0,nrow=n)
    s<-m
    # Filter loop
    for(t in 1:n)
      { if(t<(p+2))
          { m[t]<-mean(y); s[t]<-var(y); }
        else
          { ylag<-t(y[(t-p):(t-1)])%*%rev(r)
            m[t]<-ylag + (1-sum(r))*xd[t]
            s[t]<-(sum(r^2)*var(y[1:t-1]) + (1-sum(r^2))*var(xd[1:t-1]))
          }
      }
    return(ts(matrix(cbind(m,s),ncol=2),start=1,names=c("m","s2")))

  }

# parpllf -- PAR(p) log-likelihood function.
# Returns the value of the LLF for the given data and parameters.

parpllf<-function(p,y=y,x=matrix(0,nrow=length(y)))
  { # initialize the parameters
    k<-ncol(x)
    no.param<-length(p)
    r<-as.matrix(p[1:(no.param-k)])
    d<-as.matrix(p[(no.param-k+1):length(p)])
    y<-as.matrix(y); x<-as.matrix(x);
    # Call the filter
    pf<-parpfilter(y,x,r,d);
    # Extract vectors of parameters from filter output
    pf<-pf[(no.param-k+1):nrow(pf),];
    m<-pf[,"m"]; s2<-pf[,"s2"]; 
    x<-x[(no.param-k+1):nrow(x),];    y<-y[(no.param-k+1):nrow(y)];
    # Compute log-likelihood function
    lp<-(lgamma((s2*m) + y) - lgamma(s2*m) - lgamma(y+1)
         + (s2*m*log(s2)) - ((s2*m + y)*log(1+s2)))
    lp<-sum(lp)
    lp
  }

# Parp -- this is the main function for calling the PAR(p) model
# estimation.
# y = Event count time series you wish to fit.  (can be a matrix or a
# ts object)
# x = matrix of regressors,including an intercept
# init.param = starting values for the optimizer.  This is the p AR
# parameters, then the k regressors, or
# init.param = c(AR(1),AR(2),...,AR(p),d_1,d_2,...,d_k)
#

Parp<-function(y,x=matrix(1,nrow=length(y)),init.param,...)
  { y<-matrix(y,ncol=1)
    # Do a quick check for R versus S....this still needs to be debugged!
    if (exists("is.R") && is.function(is.R) && is.R()) {
    fitted.param<-optim(init.param, parpllf, method = "Nelder-Mead", hessian=TRUE, 
                        control=list(trace=2, REPORT=5, fnscale=-(length(y))),
                        y=y,x=x,...)
    param<-fitted.param$par
    llf <- fitted.param$value
    } else {
      fitted.param<-nlminb(start = init.param, obj = parpllf,
                           hessian=F, scale = -(length(y)), y=y, x=x,...)
      param<-fitted.param$parameters
    }
    
    se<-sqrt(diag(-solve(fitted.param$hessian)))
    z<-param/se
    coefs <- t(matrix(rbind(param,se,z),nrow=3));
    colnames(coefs) <- c("Parameters","Std. Errors","Z-score")
     cat(" ","\n")
     cat("PAR(p) regression output","\n")
     cat("--------------------------------------","\n")
     print.coefmat(coefs)
     cat("--------------------------------------","\n")

    fit <- list(coefs=coefs,
                   optim.param=fitted.param,
                   param=param,
                   std.err=se,
                   z=z,
                   llf=llf)
    class(fit) <- c("pests")
    invisible(fit)
  }






